import numpy as np
import pandas as pd
from random import random, randint

def get_customer(ctype=None):
    """Customers come from two feature distributions.
    Class 1: mean age 25, var 5 years, min age 18
             mean ARPU 100, var 15
    Class 2: mean age 45, var 6 years
             mean ARPU 50, var 25
    """
    if ctype is None:
        if random() > .5: #coin toss
            ctype = 1
        else:
            ctype = 2
    age = 0
    ft = -1
    if ctype == 1:
        while age < 18:
            age = np.random.normal(25, 5)
        while ft < 0:
            ft = np.random.normal(100, 15)
    if ctype == 2:
        while age < 18:
            age = np.random.normal(45, 6)
        while ft < 0:
            ft = np.random.normal(50, 25)
    age = round(age)
    return ctype, (age, ft)

def get_rewards(customer):
    """
    There are three actions:
    promo 1: low value. 10 dollar if accept
    promo 2: mid value. 25 dollar if accept
    promo 3: high value. 100 dollar if accept

    Both groups are unlikely to accept promo 2.
    Group 1 is more likely to accept promo 1.
    Group 2 is slightly more likely to accept promo 3.

    The optimal choice for group 1 is promo 1; 90% acceptance for
    an expected reward of 9 dollars each.
    Group 2 accepts with 25% rate for expected 2.5 dollar reward

    The optimal choice for group 2 is promo 3; 20% acceptance for an expected
    reward of 20 dollars each.
    Group 1 accepts with 2% for expected reward of 2 dollars.

    The least optimal choice in all cases is promo 2; 10% acceptance rate for both groups
    for an expected reward of 2.5 dollars.
    """
    if customer[0] == 1: #group 1 customer
        if random() > .1:
            reward1 = 10
        else:
            reward1 = 0
        if random() > .90:
            reward2 = 25
        else:
            reward2 = 0
        if random() > .98:
            reward3 = 100
        else:
            reward3 = 0
    if customer[0] == 2: #group 2 customer
        if random() > .75:
            reward1 = 10
        else:
            reward1 = 0
        if random() > .90:
            reward2 = 25
        else:
            reward2 = 0
        if random() > .80:
            reward3 = 100
        else:
            reward3 = 0
    return np.array([reward1, reward2, reward3])

def get_cust_reward():
    """returns a customer and reward vector"""
    cust = get_customer()
    reward = get_rewards(cust)
    fts = cust[1]
    return np.array([fts])/100, reward

def generate_dataframe(n_rows):
    df = pd.DataFrame()
    ages = []
    ARPUs = []
    actions = []
    rewards = []
    for i in range(n_rows):
        cust = get_customer()
        reward_vec = get_rewards(cust)
        context = np.array([cust[1]])
        ages.append(context[0, 0])
        ARPUs.append(context[0, 1])
        action = np.random.randint(0,3)
        actions.append(action)
        reward = reward_vec[action]
        rewards.append(reward)

    df['age'] = ages
    df['ARPU'] = ARPUs
    df['action'] = actions
    df['reward'] = rewards

    return df

def get_customer_nl(ctype=None):
    """Customers come from two feature distributions.
    Class 1: mean age 25, var 5 years, min age 18
             mean ARPU 100, var 15
    Class 2: mean age 45, var 6 years
             mean ARPU 50, var 25
    """
    if ctype is None:
        ctype = randint(0,2)
    age = 0
    ft = -1
    if ctype == 0:
        while age < 18:
            age = np.random.normal(25, 5)
            ft = 125 - .1*(age-25)*(age-25) + np.random.normal(0, 4)
    if ctype == 1:
        while age < 18:
            age = np.random.normal(35, 2)
        while ft < 0:
            ft = np.random.normal(75, 3)
    if ctype == 2:
        while age < 18:
            age = np.random.normal(45, 6)
            ft = 25 + .25*(age-45)*(age-45) + np.random.normal(0, 4)
    age = round(age)
    return ctype, (age, ft)

def get_rewards_nl(customer):
    """
    There are three actions:
    promo 1: low value. 10 dollar if accept
    promo 2: mid value. 25 dollar if accept
    promo 3: high value. 100 dollar if accept

    Expected Value Matrix:

           group1|group2|group3
    ----------------------------
    promo1|  $9  |  $1  |  $1
    ----------------------------
    promo2| $2.5 | $12.5| $1.25
    ----------------------------
    promo3| $1   |  $5  | $25

    We can see each group has an optimal choice.
    """
    if customer[0] == 0: #group 1 customer
        if random() > .1:
            reward1 = 10
        else:
            reward1 = 0
        if random() > .90:
            reward2 = 25
        else:
            reward2 = 0
        if random() > .99:
            reward3 = 100
        else:
            reward3 = 0
    if customer[0] == 1: #group 2 customer
        if random() > .9:
            reward1 = 10
        else:
            reward1 = 0
        if random() > .50:
            reward2 = 25
        else:
            reward2 = 0
        if random() > .95:
            reward3 = 100
        else:
            reward3 = 0
    if customer[0] == 2: #group 3 customer
        if random() > .9:
            reward1 = 10
        else:
            reward1 = 0
        if random() > .95:
            reward2 = 25
        else:
            reward2 = 0
        if random() > .75:
            reward3 = 100
        else:
            reward3 = 0
    return np.array([reward1, reward2, reward3])

def get_cust_reward_nl():
    """
    returns a customer and reward vector
    for nonlinear problem
    """
    cust = get_customer()
    reward = get_rewards(cust)
    age = cust[1]
    return np.array([age]), reward

def generate_biased_dataframe(n_rows):
    df = pd.DataFrame()
    ages = []
    ARPUs = []
    actions = []
    rewards = []
    for i in range(n_rows):
        cust = get_customer()
        reward_vec = get_rewards(cust)
        context = np.array([cust[1]])
        age = context[0, 0]
        ARPU = context[0, 1]
        ages.append(age)
        ARPUs.append(ARPU)
        if ARPU <= 50:
            action = 0
        elif ARPU <= 100:
            action = 1
        else:
            action = 2
        actions.append(action)
        reward = reward_vec[action]
        rewards.append(reward)

    df['age'] = ages
    df['ARPU'] = ARPUs
    df['action'] = actions
    df['reward'] = rewards

    return df



"""Contextual algorithm that keeps a full linear posterior for each arm."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

np.seterr(all='warn')

from scipy.stats import invgamma

from .bandit_algorithm import BanditAlgorithm
from .contextual_dataset import ContextualDataset
import torch

import pickle
import multiprocessing

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

#These functions help with multiprocessing random number generation.
mus = None
covs = None

def get_mn(i):
    """helper function to parallelize random number generation"""
    mu = mus[i]
    cov = covs[i]
    min_eig = np.min(np.real(np.linalg.eigvals(cov)))
    if min_eig < 0:
        cov -= 10*min_eig * np.eye(*cov.shape)
    return np.random.multivariate_normal(mu, cov)

def parallelize_multivar(mus, covs, n_threads=-1):
    """parallelizes mn computation"""
    if n_threads == -1:
        try:
            cpus = multiprocessing.cpu_count()-1
        except NotImplementedError:
            cpus = 2   # arbitrary default
    else:
        cpus = n_threads

    with multiprocessing.Pool(processes=cpus) as pool:
        samples = pool.map(get_mn, range(len(mus)))
    return samples


class LinearBandits(BanditAlgorithm):
    """Thompson Sampling with independent linear models and unknown noise var."""

    def __init__(
        self,
        num_actions,
        num_features,
        name='linear_model',
        a0=6,
        b0=6,
        lambda_prior=0.25,
        initial_pulls=2
        ):
        """
        A bayesian-linear contextual bandits model.
        Assume a linear model for each action i: reward = context^T beta_i + noise
        Each beta_i has a Gaussian prior (lambda parameter), each sigma2_i (noise
        level) has an inverse Gamma prior (a0, b0 parameters). Mean, covariance,
        and precision matrices are initialized, and the ContextualDataset created.

        num_actions (int): the number of available actions in problem

        num_features (int): the length of context vector, a.k.a. the number of features

        a0 (int): initial alpha value (default 6)

        b0 (int): initial beta_0 value (default 6)

        lambda_prior (float): lambda prior parameter(default 0.25)

        initial_pulls (int): number of pure exploration rounds before Thompson sampling
        """
        hparams = {
                    'num_actions':num_actions,
                    'context_dim':num_features,
                    'a0':a0,
                    'b0':b0,
                    'lambda_prior':lambda_prior,
                    'initial_pulls':initial_pulls
        }

        self.name = name
        self.hparams = hparams

        # Gaussian prior for each beta_i
        self._lambda_prior = self.hparams['lambda_prior']

        self.mu = [
            np.zeros(self.hparams['context_dim'] + 1)
            for _ in range(self.hparams['num_actions'])
        ]

        self.cov = [(1.0 / self.lambda_prior) * np.eye(self.hparams['context_dim'] + 1)
                    for _ in range(self.hparams['num_actions'])]

        self.precision = [
            self.lambda_prior * np.eye(self.hparams['context_dim'] + 1)
            for _ in range(self.hparams['num_actions'])
        ]

        # Inverse Gamma prior for each sigma2_i
        self._a0 = self.hparams['a0']
        self._b0 = self.hparams['b0']

        self.a = [self._a0 for _ in range(self.hparams['num_actions'])]
        self.b = [self._b0 for _ in range(self.hparams['num_actions'])]

        self.t = 0
        self.data_h = ContextualDataset(self.hparams['context_dim'],
                                        self.hparams['num_actions'],
                                        intercept=True)

    def expected_values(self, context):
        """
        Computes expected values from context. Does not consider uncertainty.
        Args:
          context: Context for which the action need to be chosen.
        Returns:
          expected reward vector.
        """
        # Compute sampled expected values, intercept is last component of beta
        vals = [
            np.dot(self.mu[i][:-1], context.T) + self.mu[i][-1]
            for i in range(self.hparams['num_actions'])
        ]
        return np.array(vals)

    def _sample(self, context, parallelize=False, n_threads=-1):
        """
        Samples beta's from posterior, and samples from expected values.
        Args:
          context: Context for which the action need to be chosen.
        Returns:
          action: sampled reward vector."""

        # Sample sigma2, and beta conditional on sigma2
        context = context.reshape(-1, self.hparams['context_dim'])
        n_rows = len(context)
        a_projected = np.repeat(np.array(self.a)[np.newaxis, :], n_rows, axis=0)
        sigma2_s = self.b * invgamma.rvs(a_projected)
        if n_rows == 1:
            sigma2_s = sigma2_s.reshape(1, -1)
        beta_s = []
        try:
            for i in range(self.hparams['num_actions']):
                global mus
                global covs
                mus = np.repeat(self.mu[i][np.newaxis, :], n_rows, axis=0)
                s2s = sigma2_s[:, i]
                rep = np.repeat(s2s[:, np.newaxis], self.hparams['context_dim']+1, axis=1)
                rep = np.repeat(rep[:, :, np.newaxis], self.hparams['context_dim']+1, axis=2)
                covs = np.repeat(self.cov[i][np.newaxis, :, :], n_rows, axis=0)
                covs = rep * covs
                if parallelize:
                    multivariates = parallelize_multivar(mus, covs, n_threads=n_threads)
                else:
                    multivariates = [np.random.multivariate_normal(mus[j], covs[j]) for j in range(n_rows)]
                beta_s.append(multivariates)
        except np.linalg.LinAlgError as e:
            # Sampling could fail if covariance is not positive definite
            # Todo: Fix This
            print('Exception when sampling from {}.'.format(self.name))
            print('Details: {} | {}.'.format(e.message, e.args))
            d = self.hparams['context_dim'] + 1
            for i in range(self.hparams['num_actions']):
                multivariates = [np.random.multivariate_normal(np.zeros((d)), np.eye(d)) for j in range(n_rows)]
                beta_s.append(multivariates)
        beta_s = np.array(beta_s)


        # Compute sampled expected values, intercept is last component of beta
        vals = [
            (beta_s[i, :, :-1] * context).sum(axis=-1) + beta_s[i, :, -1]
            for i in range(self.hparams['num_actions'])
        ]
        return np.array(vals)

    def action(self, context):
        """Samples beta's from posterior, and chooses best action accordingly.
        Args:
          context: Context for which the action need to be chosen.
        Returns:
          action: Selected action for the context.
        """

        # Round robin until each action has been selected "initial_pulls" times
        if self.t < self.hparams['num_actions'] * self.hparams['initial_pulls']:
            return self.t % self.hparams['num_actions']
        else:
            vals = self._sample(context)
            return np.argmax(vals)

    def update(self, context, action, reward):
        """Updates action posterior using the linear Bayesian regression formula.
        Args:
          context: Last observed context.
          action: Last observed action.
          reward: Last observed reward.
        """
        self.t += 1
        self.data_h.add(context, action, reward)

        self._update_action(action)

    def _update_action(self, action):
        """Updates posterior for given action"""
        # Update posterior of action with formulas: \beta | x,y ~ N(mu_q, cov_q)
        x, y = self.data_h.get_data(action)
        x = np.array(x)
        y = np.array(y)

        # The algorithm could be improved with sequential update formulas (cheaper)
        s = np.dot(x.T, x)

        # Some terms are removed as we assume prior mu_0 = 0.
        precision_a = s + self.lambda_prior * np.eye(self.hparams['context_dim'] + 1)
        cov_a = np.linalg.inv(precision_a)
        mu_a = np.dot(cov_a, np.dot(x.T, y))

        # Inverse Gamma posterior update
        a_post = self.a0 + x.shape[0] / 2.0
        b_upd = 0.5 * (np.dot(y.T, y) - np.dot(mu_a.T, np.dot(precision_a, mu_a)))
        b_post = self.b0 + b_upd

        # Store new posterior distributions
        self.mu[action] = mu_a
        self.cov[action] = cov_a
        self.precision[action] = precision_a
        self.a[action] = a_post
        self.b[action] = b_post

    def fit(self, contexts, actions, rewards, num_updates=1):
        """Inputs bulk data for training.
        Args:
          contexts: Set of observed contexts.
          actions: Corresponding list of actions.
          rewards: Corresponding list of rewards.
        """
        data_length = len(rewards)
        self.data_h._ingest_data(contexts, actions, rewards)
        self.t += data_length
        #update posterior on ingested data
        for n in range(num_updates):
            for action in range(self.data_h.num_actions):
                self._update_action(action)

    def predict(self, contexts, thompson=True, parallelize=True, n_threads=-1):
        """Takes a list or array-like of contexts and batch predicts on them"""
        try:
            contexts = contexts.values
        except:
            pass
        if thompson:
            reward_matrix = self._sample(contexts, parallelize=parallelize, n_threads=n_threads)
        else:
            reward_matrix = self.expected_values(contexts)
        return np.argmax(reward_matrix, axis=0)

    def save(self, path):
        """saves model to path"""
        with open(path, 'wb') as f:
            pickle.dump(self, f)

    @property
    def a0(self):
        return self._a0

    @property
    def b0(self):
        return self._b0

    @property
    def lambda_prior(self):
        return self._lambda_prior



"""Thompson Sampling with linear posterior over a learnt deep representation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
np.seterr(all='warn')

import warnings

from scipy.stats import invgamma

from .bandit_algorithm import BanditAlgorithm
from .contextual_dataset import ContextualDataset
from .neural_bandit_model import NeuralBanditModel
import torch

import multiprocessing
import os
import pickle
import shutil

#These functions help with multiprocessing random number generation.
mus = None
covs = None

def load_model(path):
    with open(path, 'rb') as f:
        return pickle.load(f)

def get_mn(i):
    """helper function to parallelize random number generation"""
    mu = mus[i]
    cov = covs[i]
    min_eig = np.min(np.real(np.linalg.eigvals(cov)))
    if min_eig < 0:
        cov -= 10*min_eig * np.eye(*cov.shape)
    return np.random.multivariate_normal(mu, cov)

def parallelize_multivar(mus, covs, n_threads=-1):
    """parallelizes mn computation"""
    if n_threads == -1:
        try:
            cpus = multiprocessing.cpu_count() - 1
        except NotImplementedError:
            cpus = 2   # arbitrary default
    else:
        cpus = n_threads

    with multiprocessing.Pool(processes=cpus) as pool:
        samples = pool.map(get_mn, range(len(mus)))
    return samples


class NeuralBandits(BanditAlgorithm):
    """Full Bayesian linear regression on the last layer of a deep neural net."""

    def __init__(
            self,
            num_actions,
            num_features,
            name='neural_model',
            do_scaling=True,
            init_scale=0.3,
            activation='relu',
            verbose=True,
            optimizer='RMS',
            layer_sizes=[50],
            batch_size=512,
            activate_decay=True,
            initial_lr=0.1,
            max_grad_norm=5.0,
            show_training=False,
            freq_summary=1000,
            buffer_s=-1,
            initial_pulls=100,
            reset_lr=True,
            lr_decay_rate=0.5,
            training_freq=1,
            training_freq_network=50,
            training_epochs=100,
            memory_size=-1,
            a0=6,
            b0=6,
            lambda_prior=0.25,
        ):
        """
        A "NeuralLinear" Deep Bayesian contextual bandits model.

            num_actions (int): the number of actions in problem

            num_features (int): the length of context vector, a.k.a. the number of features

            name (string): name for this model instance (default 'neural_model')

            do_scaling (bool): whether to automatically scale features (default True)

            init_scale (float): variance for neural network weights initialization (default 0.3)

            activation (string): activation function for neural network layers (default 'relu')

            verbose (bool): whether to print reports on training steps (default True)

            layer_sizes (list of integers): defines neural network architecture: n_layers = len(layer_sizes), value is per-layer width. (default [50])

            batch_size (integer): batch size for neural network training (default 512)

            activate_decay (bool): whether to use learning rate decay (default True),

            initial_lr (float): initial learning rate for neural network training (default 0.1)

            max_grad_norm (float): maximum gradient value for gradient clipping (default 5.0)

            show_training (bool): whether to show details of neural network training

            freq_summary (int): summary output frequency in number of steps (default 1000)

            buffer_s (int): buffer size for retained examples (default -1)

            initial_pulls (int): number of random pulls before greedy behavior (default 100),

            reset_lr (bool) = whether to reset learning rate on each nn training (default True),

            lr_decay_rate (float): learning rate decay for nn updates (default 0.5)

            training_freq (int): frequency for updates to bayesian linear regressor (default 1)

            training_freq_network (int): frequency of neural network re-trainings (default 50)

            training_epochs (int): number of epochs in each neural network re-training (default 100)

            a0 (int): initial alpha value (default 6)

            b0 (int): initial beta_0 value (default 6)

            lambda_prior (float): lambda prior parameter(default 0.25)
        """
        self.hparams = {
            'num_actions':num_actions,
            'context_dim':num_features,
            'name':name,
            'init_scale':init_scale,
            'activation':activation,
            'verbose':verbose,
            'optimizer':optimizer,
            'layer_sizes':layer_sizes,
            'batch_size':batch_size,
            'activate_decay':activate_decay,
            'initial_lr':initial_lr,
            'max_grad_norm':max_grad_norm,
            'show_training':show_training,
            'freq_summary':freq_summary,
            'buffer_s':buffer_s,
            'initial_pulls':initial_pulls,
            'reset_lr':reset_lr,
            'lr_decay_rate':lr_decay_rate,
            'training_freq':training_freq,
            'training_freq_network':training_freq_network,
            'training_epochs':training_epochs,
            'memory_size':memory_size,
            'a0':a0,
            'b0':b0,
            'lambda_prior':lambda_prior
        }
        self.do_scaling = do_scaling
        self.name = name
        self.latent_dim = self.hparams['layer_sizes'][-1]
        self.num_actions = self.hparams['num_actions']
        self.context_dim = self.hparams['context_dim']
        self.initial_pulls = self.hparams['initial_pulls']
        self.reset_lr = self.hparams['reset_lr']

        self.master_params = dict()

        # Gaussian prior for each beta_i
        self._lambda_prior = self.hparams['lambda_prior']

        self.mu = [
            np.zeros(self.latent_dim)
            for _ in range(self.hparams['num_actions'])
        ]

        self.cov = [(1.0 / self.lambda_prior) * np.eye(self.latent_dim)
                    for _ in range(self.hparams['num_actions'])]

        self.precision = [
            self.lambda_prior * np.eye(self.latent_dim)
            for _ in range(self.hparams['num_actions'])
        ]

        # Inverse Gamma prior for each sigma2_i
        self._a0 = self.hparams['a0']
        self._b0 = self.hparams['b0']

        self.a = [self._a0 for _ in range(self.hparams['num_actions'])]
        self.b = [self._b0 for _ in range(self.hparams['num_actions'])]

        # Regression and NN Update Frequency
        self.update_freq_lr = self.hparams['training_freq']
        self.update_freq_nn = self.hparams['training_freq_network']

        self.t = 0
        self.optimizer_n = optimizer

        self.num_epochs = self.hparams['training_epochs']

        memory_size = self.hparams['memory_size']

        self.data_h = ContextualDataset(
                            self.hparams['context_dim'],
                            self.hparams['num_actions'],
                            intercept=False,
                            memory_size=memory_size
        )

        self.latent_h = ContextualDataset(
                            self.latent_dim,
                            self.hparams['num_actions'],
                            intercept=False,
                            memory_size=memory_size
        )

        self.bnn = NeuralBanditModel(optimizer, self.hparams, f'{name}-bnn')

    def get_representation(self, context):
        """
        Returns the latent feature vector from the neural network.
        This vector is called z in the Google Brain paper.
        """
        if not isinstance(context, torch.Tensor):
            c = torch.tensor(context).float()
        else:
            c = context.float()
        z = self.bnn.get_representation(c)
        return z

    def expected_values(self, context, scale=False):
        """
        Computes expected values from context. Does not consider uncertainty.
        Args:
          context: Context for which the action need to be chosen.
        Returns:
          expected reward vector.
        """
        if scale:
            context = self.data_h.scale_contexts(contexts=context)

        # Compute last-layer representation for the current context
        z_context = self.get_representation(context).numpy()

        # Compute sampled expected values, intercept is last component of beta
        vals = [
            np.dot(self.mu[i], z_context.T)
            for i in range(self.hparams['num_actions'])
        ]
        return np.array(vals)

    def _sample(self, context, parallelize=False, n_threads=-1):
        # Sample sigma2, and beta conditional on sigma2
        n_rows = len(context)
        d = self.mu[0].shape[0]
        a_projected = np.repeat(np.array(self.a)[np.newaxis, :], n_rows, axis=0)
        sigma2_s = self.b * invgamma.rvs(a_projected)
        if n_rows == 1:
            sigma2_s = sigma2_s.reshape(1, -1)
        beta_s = []
        try:
            for i in range(self.hparams['num_actions']):
                global mus
                global covs
                mus = np.repeat(self.mu[i][np.newaxis, :], n_rows, axis=0)
                s2s = sigma2_s[:, i]
                rep = np.repeat(s2s[:, np.newaxis], d, axis=1)
                rep = np.repeat(rep[:, :, np.newaxis], d, axis=2)
                covs = np.repeat(self.cov[i][np.newaxis, :, :], n_rows, axis=0)
                covs = rep * covs
                if parallelize:
                    multivariates = parallelize_multivar(mus, covs, n_threads=n_threads)
                else:
                    multivariates = [np.random.multivariate_normal(mus[j], covs[j]) for j in range(n_rows)]
                beta_s.append(multivariates)
        except np.linalg.LinAlgError as e:
            # Sampling could fail if covariance is not positive definite
            # Todo: Fix This
            print('Exception when sampling from {}.'.format(self.name))
            print('Details: {} | {}.'.format(e.message, e.args))
            for i in range(self.hparams['num_actions']):
                multivariates = [np.random.multivariate_normal(np.zeros((d)), np.eye(d)) for j in range(n_rows)]
                beta_s.append(multivariates)
        beta_s = np.array(beta_s)

        # Compute last-layer representation for the current context
        z_context = self.get_representation(context).numpy()

        # Apply Thompson Sampling
        vals = [
            (beta_s[i, :, :] * z_context).sum(axis=-1)
            for i in range(self.hparams['num_actions'])
        ]
        return np.array(vals)

    def action(self, context):
        """Samples beta's from posterior, and chooses best action accordingly."""
        # Round robin until each action has been selected "initial_pulls" times
        if self.t < self.num_actions * self.initial_pulls:
            return self.t % self.num_actions
        else:
            context = context.reshape(-1, self.hparams['context_dim'])
            if self.do_scaling:
                context = self.data_h.scale_contexts(contexts=context)
            vals = self._sample(context)
        return np.argmax(vals)

    def update(self, context, action, reward):
        """Updates the posterior using linear bayesian regression formula."""
        self.t += 1
        self.data_h.add(context, action, reward)
        c = context.reshape((1, self.context_dim))
        z_context = self.get_representation(c)
        self.latent_h.add(z_context, action, reward)
        # Retrain the network on the original data (data_h)
        if self.t % self.update_freq_nn == 0:
            self._retrain_nn()

        if self.t % self.update_freq_lr == 0:
            self._update_actions()

    def _retrain_nn(self):
        """Retrain the network on original data (data_h)"""
        if self.reset_lr:
            self.bnn.assign_lr()

        #uncomment following lines for better stuff.
        #steps = round(self.num_epochs * (len(self.data_h)/self.hparams['batch_size']))
        #print(f"training for {steps} steps.")
        self.bnn.train(self.data_h, self.num_epochs)
        self._replace_latent_h()

    def _replace_latent_h(self):
        # Update the latent representation of every datapoint collected so far
        if self.do_scaling:
            self.data_h.scale_contexts()
        ctx = self.data_h.get_contexts(scaled=self.do_scaling)
        new_z = self.get_representation(ctx)

        self.latent_h._replace_data(contexts=new_z)

    def _update_actions(self):
        """
        Update the Bayesian Linear Regression on
        stored latent variables.
        """

        # Find all the actions to update
        actions_to_update = self.latent_h.actions

        for action_v in np.unique(actions_to_update):

            # Update action posterior with formulas: \beta | z,y ~ N(mu_q, cov_q)
            z, y = self.latent_h.get_data(action_v)
            z = z.numpy()
            y = y.numpy()

            # The algorithm could be improved with sequential formulas (cheaper)
            s = np.dot(z.T, z)

            # Some terms are removed as we assume prior mu_0 = 0.
            precision_a = s + self.lambda_prior * np.eye(self.latent_dim)
            cov_a = np.linalg.inv(precision_a)
            mu_a = np.dot(cov_a, np.dot(z.T, y))

            # Inverse Gamma posterior update
            a_post = self.a0 + z.shape[0] / 2.0
            b_upd = 0.5 * np.dot(y.T, y)
            b_upd -= 0.5 * np.dot(mu_a.T, np.dot(precision_a, mu_a))
            b_post = self.b0 + b_upd

            # Store new posterior distributions
            self.mu[action_v] = mu_a
            self.cov[action_v] = cov_a
            self.precision[action_v] = precision_a
            self.a[action_v] = a_post
            self.b[action_v] = b_post

    def fit(self, contexts, actions, rewards):
        """Inputs bulk data for training.
        Args:
          contexts: Set of observed contexts.
          actions: Corresponding list of actions.
          rewards: Corresponding list of rewards.
        """
        data_length = len(rewards)
        self.data_h._ingest_data(contexts, actions, rewards)
        #create latent representations of data
        self._replace_latent_h()
        self.latent_h.actions = self.data_h.actions
        self.latent_h.rewards = self.data_h.rewards
        #update count
        self.t += data_length
        #update posterior on ingested data
        self._retrain_nn()
        self._update_actions()


    def predict(self, contexts, thompson=True, parallelize=True, n_threads=-1):
        """Takes a list or array-like of contexts and batch predicts on them"""
        contexts = contexts.reshape(-1, self.hparams['context_dim'])
        if self.do_scaling:
            contexts = self.data_h.scale_contexts(contexts=contexts)

        if thompson:
            reward_matrix = self._sample(contexts, parallelize=parallelize, n_threads=n_threads)
        else:
            reward_matrix = self.expected_values(contexts)
        return np.argmax(reward_matrix, axis=0)

    def save(self, path):
        """saves model to path"""
        with open(path, 'wb') as f:
            pickle.dump(self, f)

    @property
    def a0(self):
        return self._a0

    @property
    def b0(self):
        return self._b0

    @property
    def lambda_prior(self):
        return self._lambda_prior



"""Define a data buffer for contextual bandit algorithms."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import numpy as np
import torch
import pandas as pd
from scipy.sparse import coo_matrix

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

class ContextualDataset(object):
    """The buffer is able to append new data, and sample random minibatches."""

    def __init__(self, context_dim, num_actions, memory_size=-1, intercept=False):
        """Creates a ContextualDataset object.
        The data is stored in attributes: contexts and rewards.
        The sequence of taken actions are stored in attribute actions.
        Args:
          context_dim: Dimension of the contexts.
          num_actions: Number of arms for the multi-armed bandit.
          memory_size: Specify the number of examples to store in memory.
            if memory_size = -1, all data will be stored.
          intercept: If True, it adds a constant (1.0) dimension to each context X,
            at the end.
        """

        self._context_dim = context_dim
        self._num_actions = num_actions
        self.contexts = None
        self.scaled_contexts = None
        self.rewards = None
        self.actions = []
        self.memory_size = memory_size
        self.intercept = intercept
        self.scaling_data = []

    def add(self, context, action, reward):
        """Adds a new triplet (context, action, reward) to the dataset.
        The reward for the actions that weren't played is assumed to be zero.
        Args:
          context: A d-dimensional vector with the context.
          action: Integer between 0 and k-1 representing the chosen arm.
          reward: Real number representing the reward for the (context, action).
        """
        if not isinstance(context, torch.Tensor):
            context = torch.tensor(context)
        if len(context.shape) > 1:
            context = context.reshape(-1)
        if self.intercept:
            c = context[:]
            c = torch.cat((c, torch.tensor([1.0]).double()))
            c = c.reshape((1, self.context_dim + 1))
        else:
            if type(context) == type(torch.tensor(0)):
                c = context[:].reshape((1, self.context_dim))
            else:
                c = torch.tensor(context[:]).reshape((1, self.context_dim))

        if self.contexts is None:
            self.contexts = c
        else:
            self.contexts = torch.cat((self.contexts, c))

        r = torch.zeros((1, self.num_actions))
        r[0, action] = float(reward)
        if self.rewards is None:
            self.rewards = r
        else:
            self.rewards = torch.cat((self.rewards, r))

        self.actions.append(action)

        #Drop oldest example if memory constraint
        if self.memory_size != -1:
            if self.contexts.shape[0] > self.memory_size:
                self.contexts = self.contexts[1:, :]
                self.rewards = self.rewards[1:, :]
                self.actions = self.actions[1:]
            #Assert lengths match
            assert len(self.actions) == len(self.rewards)
            assert len(self.actions) == len(self.contexts)

    def _replace_data(self, contexts=None, actions=None, rewards=None):
        if contexts is not None:
            self.contexts = contexts
        if actions is not None:
            self.actions = actions
        if rewards is not None:
            self.rewards = rewards

    def _ingest_data(self, contexts, actions, rewards):
        """Ingests bulk data."""
        if isinstance(rewards, pd.DataFrame) or isinstance(rewards, pd.Series):
            rewards = rewards.values
        if isinstance(actions, pd.DataFrame) or isinstance(actions, pd.Series):
            actions = actions.values
        if isinstance(contexts, pd.DataFrame) or isinstance(contexts, pd.Series):
            contexts = contexts.values
        if not isinstance(rewards, torch.Tensor):
            rewards = torch.tensor(rewards)
        if not isinstance(actions, torch.Tensor):
            actions = torch.tensor(actions)
        if not isinstance(contexts, torch.Tensor):
            contexts = torch.tensor(contexts)
        data_length = len(rewards)
        if self.memory_size != -1:
            if data_length + len(self.rewards) > self.memory_size:
                print('Cannot add more examples: ')
                raise Exception('Too many examples for specified memory_size.')
        try:
            contexts = contexts.reshape(-1, self.context_dim)
        except:
            print('Got bad data contexts shape: ', contexts.shape)
            raise Exception('Expected: ({}, {})'.format(data_length, self.context_dim))
        if self.intercept:
            #add intercepts
            contexts = torch.cat([contexts, torch.ones((data_length, 1)).double()], dim=1)
        try:
            assert len(contexts) == data_length
            assert len(actions) == data_length
        except AssertionError:
            raise AssertionError('Data lengths inconsistent.')
        if self.contexts is None:
            self.contexts = contexts
        else:
            self.contexts = torch.cat((self.contexts, contexts))

        rewards_array = coo_matrix((np.array(rewards), (np.arange(data_length), np.array(actions)))).toarray()
        rewards_array = torch.tensor(rewards_array)
        if self.rewards is None:
            self.rewards = rewards_array
        else:
            self.rewards = torch.cat((self.rewards, rewards_array.float()))

        self.actions = self.actions + list(actions)

    def get_batch(self, batch_size=512):
        """Returns a random minibatch of (contexts, rewards) with batch_size."""
        n, _ = self.contexts.shape
        ind = np.random.choice(range(n), batch_size)
        return self.contexts[ind, :], self.rewards[ind, :], torch.tensor(self.actions)[ind]

    def get_data(self, action):
        """Returns all (context, reward) where the action was played."""
        n, _ = self.contexts.shape
        ind = np.argwhere(np.array(self.actions) == action).reshape(-1)
        return self.contexts[ind, :], self.rewards[ind, action]

    def get_data_with_weights(self):
        """Returns all observations with one-hot weights for actions."""
        weights = torch.zeros((self.contexts.shape[0], self.num_actions))
        a_ind = np.array([(i, val) for i, val in enumerate(self.actions)])
        weights[a_ind[:, 0], a_ind[:, 1]] = 1.0
        return self.contexts, self.rewards, weights

    def get_batch_with_weights(self, batch_size, scaled=False):
        """Returns a random mini-batch with one-hot weights for actions."""

        n, _ = self.contexts.shape
        if self.memory_size == -1:
            # use all the data
            ind = np.random.choice(range(n), batch_size)
        else:
            # use only buffer (last buffer_s obs)
            ind = np.random.choice(range(max(0, n - self.memory_size), n), batch_size)

        weights = torch.zeros((batch_size, self.num_actions))
        sampled_actions = torch.tensor(self.actions)[ind]
        a_ind = torch.tensor([(i, val) for i, val in enumerate(sampled_actions)])
        weights[a_ind[:, 0], a_ind[:, 1]] = 1.0
        if scaled:
            ctx = self.scaled_contexts[ind, :]
        else:
            ctx = self.contexts[ind, :]
        return ctx, self.rewards[ind, :], weights

    def num_points(self, f=None):
        """Returns number of points in the buffer (after applying function f)."""
        if f is not None:
            return f(self.contexts.shape[0])
        return self.contexts.shape[0]

    def scale_contexts(self, contexts=None):
        """
        Performs mean/std scaling operation on contexts.
        if contexts is provided as argument, returns scaled version
            (scaled by statistics of data in dataset.)
        """
        means = self.contexts.mean(dim=0)
        stds = self.contexts.std(dim=0)
        stds[stds==0] = 1
        self.scaled_contexts = self.contexts.clone()
        for col in range(self._context_dim):
            self.scaled_contexts[:, col] -= means[col]
            self.scaled_contexts[:, col] /= stds[col]
        if contexts is not None:
            if not isinstance(contexts, torch.Tensor):
                contexts = torch.tensor(contexts)
            result = contexts
            for col in range(self._context_dim):
                result[:, col] -= means[col]
                result[:, col] /= stds[col]
            return result

    def get_contexts(self, scaled=False):
        if scaled:
            return self.scaled_contexts
        else:
            return self.contexts

    def __len__(self):
        return len(self.actions)

    @property
    def context_dim(self):
        return self._context_dim

    @property
    def num_actions(self):
        return self._num_actions

    @property
    def contexts(self):
        return self._contexts

    @contexts.setter
    def contexts(self, value):
        self._contexts = value

    @property
    def actions(self):
        return self._actions

    @actions.setter
    def actions(self, value):
        self._actions = value

    @property
    def rewards(self):
        return self._rewards

    @rewards.setter
    def rewards(self, value):
        self._rewards = value



from distutils.core import setup
import os

long_desc = """
A practical library for building contextual bandits models with deep Bayesian approximation.
Supports both online learning and offline training of models as well as novel methods for cross-validating CB models on historic data.
"""

def parse_requirements():
    """reads dependencies in from requirements.txt"""
    setup_dir = os.path.dirname(os.path.realpath(__file__))
    requirements_path = os.path.join(setup_dir, 'requirements.txt')
    with open(requirements_path, 'r') as f:
        requirements = f.read()
    return requirements.split('\n')

def parse_version():
    """gets current version of library"""
    setup_dir = os.path.dirname(os.path.realpath(__file__))
    version_path = os.path.join(setup_dir, 'space_bandits', '__init__.py')
    with open(version_path, 'r') as f:
        content = f.read()
    lines = content.split('\n')
    for line in lines:
        if "__version__" in line:
            version = line.split('=')[1]
    return version.strip()[1:-1]

version = parse_version()

setup(
    name='space-bandits',
    version=f'{version}',
    description='Deep Bayesian Contextual Bandits Library',
    long_description=long_desc,
    author='Michael Klear',
    author_email='michael@fellowship.ai',
    url='https://github.com/fellowship/space-bandits',
    download_url=f'https://github.com/fellowship/space-bandits/archive/v{version}.tar.gz',
    install_requires=parse_requirements(),
    packages=['space_bandits']
)



"""Define a family of neural network architectures for bandits.
The network accepts different type of optimizers that could lead to different
approximations of the posterior distribution or simply to point estimates.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import torch
from torch import nn
import torch.nn.functional as F

from .bayesian_nn import BayesianNN

def build_action_mask(actions, num_actions):
    """
    Takes a torch tensor with integer values.
    Returns a one-hot encoded version, where
        each column corresponds to a single action.
    """
    ohe = torch.zeros((len(actions), num_actions))
    actions = actions.reshape(-1, 1)
    return ohe.scatter_(1, actions, 1)

def build_target(rewards, actions, num_actions):
    """
    Takes a torch tensor with floating point values.
    Returns a one-hot encoded version, where
        each column corresponds to a single action.
        The value is the observed reward.
    """
    ohe = torch.zeros((len(actions), num_actions))
    actions = actions.reshape(-1, 1)
    return ohe.scatter_(1, actions, rewards)


class NeuralBanditModel(nn.Module):
    """Implements a neural network for bandit problems."""

    def __init__(self, optimizer, hparams, name):
        """Saves hyper-params and builds a torch NN."""
        super(NeuralBanditModel, self).__init__()
        self.opt_name = optimizer
        self.name = name
        self.hparams = hparams
        self.verbose = self.hparams["verbose"]
        self.times_trained = 0
        self.lr = self.hparams["initial_lr"]
        if self.hparams['activation'] == 'relu':
            self.activation = F.relu
        else:
            act = self.hparams['activation']
            msg = f'activation {act} not implimented'
            raise Exception(msg)
        self.layers = []
        self.build_model()
        self.optim = self.select_optimizer()
        self.loss = nn.modules.loss.MSELoss()

    def build_layer(self, inp_dim, out_dim):
        """Builds a layer with input x; dropout and layer norm if specified."""

        init_s = self.hparams.get('init_scale', 0.3)
        #these features not currently implemented
        layer_n = self.hparams.get("layer_norm", False)
        dropout = self.hparams.get("use_dropout", False)

        layer = nn.modules.linear.Linear(
            inp_dim,
            out_dim
        )
        nn.init.uniform_(layer.weight, a=-init_s, b=init_s)
        name = f'layer {len(self.layers)}'
        self.add_module(name, layer)
        return layer

    def forward(self, x):
        """forward pass of the neural network"""
        for i, layer in enumerate(self.layers):
            x = layer(x)
            if i != len(self.layers)-1:
                x = self.activation(x)
        return x

    def build_model(self):
        """
        Defines the actual NN model with fully connected layers.
        """
        for i, layer in enumerate(self.hparams['layer_sizes']):
            if i==0:
                inp_dim = self.hparams['context_dim']
            else:
                inp_dim = self.hparams['layer_sizes'][i-1]
            out_dim = self.hparams['layer_sizes'][i]
            new_layer = self.build_layer(inp_dim, out_dim)
            self.layers.append(new_layer)
        output_layer = nn.modules.linear.Linear(out_dim, self.hparams['num_actions'])
        self.layers.append(output_layer)

    def assign_lr(self, lr=None):
        """
        Resets the learning rate to input argument value "lr".
        """
        if lr is None:
            lr = self.lr
        for param_group in self.optim.param_groups:
            param_group['lr'] = lr

    def select_optimizer(self):
        """Selects optimizer. To be extended (SGLD, KFAC, etc)."""
        lr = self.hparams['initial_lr']
        return torch.optim.RMSprop(self.parameters(), lr=lr)

    def scale_weights(self):
        init_s = self.hparams.get('init_scale', 0.3)
        for layer in self.layers:
            nn.init.uniform_(layer.weight, a=-init_s, b=init_s)

    def do_step(self, x, y, w, step):

        decay_rate = self.hparams.get('lr_decay_rate', 0.5)
        base_lr = self.hparams.get('initial_lr', 0.1)

        lr = base_lr * (1 / (1 + (decay_rate * step)))
        self.assign_lr(lr)

        y_hat = self.forward(x.float())
        y_hat *= w
        ls = self.loss(y_hat, y.float())
        ls.backward()
        clip = self.hparams['max_grad_norm']
        torch.nn.utils.clip_grad_norm_(self.parameters(), clip)

        self.optim.step()
        self.optim.zero_grad()

    def train(self, data, num_steps):
        """Trains the network for num_steps, using the provided data.
        Args:
          data: ContextualDataset object that provides the data.
          num_steps: Number of minibatches to train the network for.
        """

        if self.verbose:
            print("Training {} for {} steps...".format(self.name, num_steps))

        batch_size = self.hparams.get('batch_size', 512)

        data.scale_contexts()
        for step in range(num_steps):
            x, y, w = data.get_batch_with_weights(batch_size, scaled=True)
            self.do_step(x, y, w, step)

    def get_representation(self, contexts):
        """
        Given input contexts, returns representation
        "z" vector.
        """
        x = contexts
        with torch.no_grad():
            for i, layer in enumerate(self.layers[:-1]):
                x = layer(x)
                x = self.activation(x)
        return x
